"""GroundOverlay module."""

from abc import ABC
from typing import Any, Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Icon import Icon
from pyLiveKML.objects.Object import (
    _BaseObject,
    _ChildDef,
    _FieldDef,
    _KMLDump,
    _KMLParser,
    _Angle180,
    _Angle90,
    Object,
)
from pyLiveKML.objects.Overlay import Overlay
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.types import AltitudeModeEnum, GeoColor


class LatLonBox(Object):
    """A `<LatLonBox>` KML tag constructor.

    Specific to the `<GroundOverlay>` KML tag. Specifies where the top, bottom, right,
    and left sides of a bounding box for the ground overlay are aligned.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-groundoverlay

    Parameters
    ----------
    north : float
        Specifies the latitude of the north edge of the bounding box, in decimal degrees
        from 0 to ±90.
    south : float
        Specifies the latitude of the south edge of the bounding box, in decimal degrees
        from 0 to ±90.
    east : float
        Specifies the longitude of the east edge of the bounding box, in decimal degrees
        from 0 to ±180. (For overlays that overlap the meridian of 180° longitude, values
        can extend beyond that range.)
    west : float
        Specifies the longitude of the west edge of the bounding box, in decimal degrees
        from 0 to ±180. (For overlays that overlap the meridian of 180° longitude, values
        can extend beyond that range.)
    rotation : float, default = 0
        Specifies a rotation of the overlay about its center, in degrees. Values can be
        ±180. The default is 0 (north). Rotations are specified in a counterclockwise
        direction.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "LatLonBox"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("north", parser=_Angle90),
        _FieldDef("south", parser=_Angle90),
        _FieldDef("east", parser=_Angle180),
        _FieldDef("west", parser=_Angle180),
        _FieldDef("rotation", parser=_Angle180),
    )
    _suppress_id = True

    def __init__(
        self,
        north: float,
        south: float,
        east: float,
        west: float,
        rotation: float = 0,
    ) -> None:
        """LatLonBox instance constructor."""
        super().__init__()
        self.north = north
        self.south = south
        self.east = east
        self.west = west
        self.rotation = rotation


class _QuadParser(_KMLParser):
    """Custom parser for LatLonQuad coordinates field."""

    @classmethod
    def parse(cls, value: Any) -> Any:
        return (
            (_Angle180.parse(value[0][0]), _Angle90.parse(value[0][1])),
            (_Angle180.parse(value[1][0]), _Angle90.parse(value[1][1])),
            (_Angle180.parse(value[2][0]), _Angle90.parse(value[2][1])),
            (_Angle180.parse(value[3][0]), _Angle90.parse(value[3][1])),
        )


class _QuadDump(_KMLDump):
    """Custom dumper for LatLonQuad coordinates field."""

    @classmethod
    def dump(cls, value: Any) -> Any:
        return " ".join((",".join(map(str, x)) for x in value))


class LatLonQuad(Object):
    """A `<LatLonQuad>` KML tag constructor.

    Specific to the `<GroundOverlay>` KML tag. Allows nonrectangular quadrilateral ground
    overlays.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#gx:latlonquad

    Parameters
    ----------
    coordinates : tuple[tuple[float, float], tuple[float, float], tuple[float, float], tuple[float, float]]
        Specifies the coordinates of the four corner points of a quadrilateral defining
        the overlay area. Exactly four coordinate tuples have to be provided, each
        consisting of floating point values for longitude and latitude. Insert a space
        between tuples. Do not include spaces within a tuple. The coordinates must be
        specified in counter-clockwise order with the first coordinate corresponding to
        the lower-left corner of the overlayed image. The shape described by these
        corners must be convex.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:LatLonQuad"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("coordinates", parser=_QuadParser, dumper=_QuadDump),
    )
    _suppress_id = True

    def __init__(
        self,
        coordinates: tuple[
            tuple[float, float],
            tuple[float, float],
            tuple[float, float],
            tuple[float, float],
        ],
    ) -> None:
        """LatLonQuad instance constructor."""
        super().__init__()
        self.coordinates = coordinates


class GroundOverlay(Overlay):
    """A KML `<GroundOverlay>` tag constructor.

    This element draws an image overlay draped onto the terrain. The `<href>` child of
    `<Icon>` specifies the image to be used as the overlay. This file can be either on
    a local file system or on a web server. If this element is omitted or contains no
    `<href>`, a rectangle is drawn using the color and `<LatLonBox>` bounds defined by
    the ground overlay.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#groundoverlay

    Parameters
    ----------
    quad : LatLonQuad | None, default = None
        A non-rectilinear bounding box definition
    box : LatLonBox | None, default = None
        A rectilinear bounding box definition
    icon : str | Icon | None, default = None
        Defines the location of the image to be used as the `Overlay`. This location can
        be either on a local file system or on a web server. If this attribute is `None`,
        a rectangle is drawn using the color and size defined by the ground or screen
        overlay. If a simple `str` is supplied, then it will be used as the `href` of an
        `Icon`.
    draw_order : int | None, default = None
        Defines the stacking order for the images in overlapping overlays. Overlays with
        higher `draw_order` values are drawn on top of overlays with lower `draw_order`
        values.
    color : GeoColor | int | None, default = None
        The color of the overlay.
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "GroundOverlay"
    _kml_fields = Overlay._kml_fields + (
        _FieldDef("altitude"),
        _FieldDef("altitude_mode", "gx:altitudeMode"),
    )
    _kml_children = Overlay._kml_children + (
        _ChildDef("box"),
        _ChildDef("quad"),
    )

    def __init__(
        self,
        # GroundOverlay parameters
        altitude: float | None = None,
        altitude_mode: AltitudeModeEnum | None = None,
        quad: LatLonQuad | None = None,
        box: LatLonBox | None = None,
        # Overlay parameters
        icon: str | Icon | None = None,
        draw_order: int | None = None,
        color: GeoColor | int | None = None,
        # Feature parameters
        name: str | None = None,
        visibility: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
    ):
        """IconStyle instance constructor."""
        Overlay.__init__(
            self,
            icon=icon,
            draw_order=draw_order,
            color=color,
            name=name,
            visibility=visibility,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
        )
        self.altitude = altitude
        self.altitude_mode = altitude_mode
        self.quad = quad
        self.box = box



"""ExtendedData module."""

from abc import ABC
from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import (
    _BaseObject,
    _DependentDef,
    _FieldDef,
)


class _ExtendedDataItem(_BaseObject, ABC):
    """Private base class for `ExtendedData` items."""

    def __init__(self) -> None:
        """_ExtendedDataItem instance constructor."""
        _BaseObject.__init__(self)
        ABC.__init__(self)


class DataItem(_ExtendedDataItem):
    """A KML `Data` tag constructor.

    Used only by `ExtendedData`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#extendeddata
    * https://developers.google.com/kml/documentation/extendeddata

    Parameters
    ----------
    name : str
    display_name : str | None, default = None
    value : str | None, default = None

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "Data"
    _kml_fields = _BaseObject._kml_fields + (
        _FieldDef("display_name", "displayName"),
        _FieldDef("value"),
    )

    def __init__(
        self,
        name: str,
        display_name: str | None,
        value: str | None,
    ) -> None:
        """DataItem instance constructor."""
        super().__init__()
        self.name = name
        self.display_name = display_name
        self.value = value

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Override `construct_kml` for fine control."""
        root = etree.Element(self._kml_tag, attrib={"name": self.name})
        self.build_kml(root, with_children, with_dependents)
        return root


class SchemaDataItem(_ExtendedDataItem):
    """A KML `SchemaData` tag constructor.

    Used only by `ExtendedData`.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#extendeddata
    * https://developers.google.com/kml/documentation/extendeddata

    Parameters
    ----------
    schema_ref : str
        A # reference to a `Schema` `id` in the hosting `Document`.
    data : dict[str, str]
        For each key-value pair in `data`:
        * The key must be a `SimpleField` of that name in the `Schema` referenced by
        `schema_ref`.
        * The value will be displayed using the `SimpleField` reference by the key.

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "SchemaData"

    def __init__(
        self,
        schema_ref: str,
        data: dict[str, str],
    ) -> None:
        """SchemaDataItem instance constructor."""
        super().__init__()
        self.schema_ref = schema_ref
        self.data = data

    def construct_kml(
        self, with_children: bool = True, with_dependents: bool = True
    ) -> etree.Element:
        """Override `construct_kml` for fine control."""
        root = etree.Element(self._kml_tag, attrib={"schemaUrl": self.schema_ref})
        self.build_kml(root, with_children, with_dependents)
        return root

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Override `build_kml` for fine control."""
        for k, v in self.data.items():
            etree.SubElement(root, "SimpleData", attrib={"name": k}).text = v


class ExtendedData(_BaseObject):
    """A KML `<ExtendedData>` tag constructor.

    The `ExtendedData` class offers three techniques for adding custom data to a KML
    `Feature`. These techniques are:

    * Adding untyped data/value pairs using `DataItem` instances (basic).
    * Declaring new typed fields using `Schema` and then instancing them using
    `SchemaDataItem` instances (advanced).
    * Referring to XML elements defined in other namespaces by referencing the external
    namespace within the KML file (basic). **This method is *not* supported by pyLiveKML**.

    These techniques can be combined within a single KML file or `Feature` for different
    pieces of data.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#extendeddata
    * https://developers.google.com/kml/documentation/extendeddata

    Parameters
    ----------
    items : _ExtendedDataItem | Iterable[_ExtendedDataItem] | None, default = None
        The items to be included in the `ExtendedData`. In practice, these may be any
        combination of `DataItem` and `SchemaDataItem` instances, in any order.

    Attributes
    ----------
    Same as parameters

    """

    _kml_tag = "ExtendedData"
    _kml_dependents = _BaseObject._kml_dependents + (_DependentDef("items"),)

    def __init__(
        self,
        items: _ExtendedDataItem | Iterable[_ExtendedDataItem] | None = None,
    ) -> None:
        """Construct Schema instances."""
        super().__init__()
        self._items = list[_ExtendedDataItem]()
        self.items = items

    @property
    def items(self) -> Iterator[_ExtendedDataItem]:
        """Generator over items."""
        yield from self._items

    @items.setter
    def items(
        self, value: _ExtendedDataItem | Iterable[_ExtendedDataItem] | None
    ) -> None:
        self._items.clear()
        if value is not None:
            if isinstance(value, _ExtendedDataItem):
                self._items.append(value)
            else:
                self._items.extend(value)



"""SoundCue module."""

from lxml import etree  # type: ignore

from pyLiveKML.objects.Object import _FieldDef
from pyLiveKML.objects.TourPrimitive import TourPrimitive


class SoundCue(TourPrimitive):
    """A KML `<gx:SoundCue>` tag constructor.

    Notes
    -----
    There is no play duration. The sound file plays in parallel to the rest of the tour,
    meaning that the next tour primitive takes place immediately after the `SoundCue`
    tour primitive is reached. If another sound file is cued before the first has
    finished playing, the files are mixed.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#extended-by_8

    Parameters
    ----------
    href: str
        Specifies a sound file to play, in MP3, M4A, or AAC format.
    delayed_start: float, default = 0
        Specifies to delay the start of the sound for a given number of seconds before
        playing the file.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "gx:SoundCue"
    _kml_fields = TourPrimitive._kml_fields + (
        _FieldDef("href"),
        _FieldDef("delayed_start", "gx:delayedStart"),
    )

    def __init__(
        self,
        href: str,
        delayed_start: float = 0,
    ) -> None:
        """SoundCue instance constructor."""
        TourPrimitive.__init__(self)
        self.href = href
        self.delayed_start = delayed_start



"""Style module."""

from typing import Any

from lxml import etree  # type: ignore

from pyLiveKML.objects.BalloonStyle import BalloonStyle
from pyLiveKML.objects.IconStyle import IconStyle
from pyLiveKML.objects.LabelStyle import LabelStyle
from pyLiveKML.objects.LineStyle import LineStyle
from pyLiveKML.objects.ListStyle import ListStyle
from pyLiveKML.objects.Object import _ChildDef
from pyLiveKML.objects.PolyStyle import PolyStyle
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.SubStyle import SubStyle


class Style(StyleSelector):
    """A KML `<Style>` tag constructor.

    A `Style` defines an addressable style group that can be referenced by `StyleMap`s
    and `Feature`s. `Style`s affect how `Geometry` is presented in the 3D viewer and how
    `Feature`s appear in the "Places" panel of the List view. Shared styles are collected
    in a `Document` and must have an `id` defined for them so that they can be referenced
    by the individual `Feature`s that use them.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#style

    Parameters
    ----------
    *args | SubStyle
        Any number of positional `SubStyle` arguments. They will be assigned to the
        attribute of the correct type. There is only one instance of each attribute, so
        duplicates will overwrite any previous attribute value.
    balloon_style : BalloonStyle | None, default = None
    icon_style : IconStyle | None, default = None
    label_style : LabelStyle | None, default = None
    line_style : LineStyle | None, default = None
    list_style : ListStyle | None, default = None
    poly_style : PolyStyle | None, default = None

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "Style"
    _kml_children = StyleSelector._kml_children + (
        _ChildDef("balloon_style"),
        _ChildDef("icon_style"),
        _ChildDef("label_style"),
        _ChildDef("line_style"),
        _ChildDef("list_style"),
        _ChildDef("poly_style"),
    )

    def __init__(
        self,
        *args: SubStyle,
        balloon_style: BalloonStyle | None = None,
        icon_style: IconStyle | None = None,
        label_style: LabelStyle | None = None,
        line_style: LineStyle | None = None,
        list_style: ListStyle | None = None,
        poly_style: PolyStyle | None = None,
    ):
        """Style instance constructor."""
        StyleSelector.__init__(self)
        self.balloon_style: BalloonStyle | None = None
        self.icon_style: IconStyle | None = None
        self.label_style: LabelStyle | None = None
        self.line_style: LineStyle | None = None
        self.list_style: ListStyle | None = None
        self.poly_style: PolyStyle | None = None
        for a in args:
            if isinstance(a, BalloonStyle):
                self.balloon_style = a
            elif isinstance(a, IconStyle):
                self.icon_style = a
            elif isinstance(a, LabelStyle):
                self.label_style = a
            elif isinstance(a, LineStyle):
                self.line_style = a
            elif isinstance(a, ListStyle):
                self.list_style = a
            elif isinstance(a, PolyStyle):
                self.poly_style = a
        if balloon_style:
            self.balloon_style = balloon_style
        if icon_style:
            self.icon_style = icon_style
        if label_style:
            self.label_style = label_style
        if line_style:
            self.line_style = line_style
        if list_style:
            self.list_style = list_style
        if poly_style:
            self.poly_style = poly_style



"""Feature module."""

from abc import ABC
from typing import Iterable, Iterator

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Object import (
    _ChildDef,
    _DependentDef,
    _FieldDef,
    _NoDump,
    Object,
)
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive
from pyLiveKML.utils import with_ns


class Feature(Object, ABC):
    """A KML `<Feature>` tag constructor.

    This is an abstract element and cannot be used directly in a KML file.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#feature

    Parameters
    ----------
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_fields = Object._kml_fields + (
        _FieldDef("name"),
        _FieldDef("is_open", "open"),
        _FieldDef("visibility"),
        _FieldDef("author_name", dumper=_NoDump),
        _FieldDef("author_link", dumper=_NoDump),
        _FieldDef("address"),
        _FieldDef("snippet", dumper=_NoDump),
        _FieldDef("snippet_max_line", dumper=_NoDump),
        _FieldDef("phone_number", "phoneNumber"),
        _FieldDef("description"),
        _FieldDef("style_url", "styleUrl"),
    )
    _kml_children: tuple[_ChildDef, ...] = Object._kml_children + (
        _ChildDef("abstract_view"),
        _ChildDef("region"),
        _ChildDef("styles"),
    )
    _kml_dependents = Object._kml_dependents + (
        _DependentDef("time_primitive"),
        _DependentDef("extended_data"),
    )

    def __init__(
        self,
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
    ):
        """Feature instance constructor."""
        Object.__init__(self)
        ABC.__init__(self)
        self.name = name
        self.is_open = is_open
        self.visibility = visibility
        self.author_name = author_name
        self.author_link = author_link
        self.address = address
        self.phone_number = phone_number
        self.snippet = snippet
        self.snippet_max_lines = snippet_max_lines
        self.description = description
        self.abstract_view = abstract_view
        self.time_primitive = time_primitive
        self.style_url = style_url
        self._styles = list[StyleSelector]()
        self.styles = styles
        self.region = region
        self.extended_data = extended_data

    @property
    def styles(self) -> Iterator[StyleSelector]:
        """Retrieve a generator over the `StyleSelector` instances in this `Feature`.

        If the property setter is called, replaces the current list of contained
        `StyleSelector`'s with those provided.

        Parameters
        ----------
        value : StyleSelector | Iterable[StyleSelector] | None
            The new `StyleSelector` elements for the `Feature`.

        :returns: A generator over the `StyleSelector`s in the `Feature`.
        :rtype: Iterator[StyleSelector]

        """
        for s in self._styles:
            yield s

    @styles.setter
    def styles(self, value: StyleSelector | Iterable[StyleSelector] | None) -> None:
        self._styles.clear()
        if value is not None:
            if isinstance(value, StyleSelector):
                self._styles.append(value)
            else:
                self._styles.extend(value)

    def build_kml(
        self,
        root: etree.Element,
        with_children: bool = True,
        with_dependents: bool = True,
    ) -> None:
        """Build the KML sub-tags for this `Feature` and append it to the provided `etree.Element`.

        Overridden from :class:`pyLiveKML.objects.Object.Object` to perform some
        additional build steps.

        Parameters
        ----------
        root : etree.Element
            The tag into which the sub-tags are to be inserted.
        with_children : bool, default = True
            Whether the `children` of the `Feature` should be constructed as sub-tags.
        with_dependents : bool, default = True
            Whether the `dependents` of the `Feature` should be constructed as sub-tags.

        """
        super().build_kml(root, with_children, with_dependents)
        if self.author_name is not None:
            author = etree.SubElement(root, with_ns("atom:author"))
            etree.SubElement(author, with_ns("atom:name")).text = self.author_name
        if self.author_link is not None:
            etree.SubElement(
                root, with_ns("atom:link"), attrib={"href": self.author_link}
            )
        if self.snippet is not None:
            attribs = {}
            if self.snippet_max_lines is not None:
                attribs["maxLines"] = str(self.snippet_max_lines)
            etree.SubElement(root, "Snippet", attribs).text = self.snippet

    def __str__(self) -> str:
        """Return a string representation."""
        return f"{self.kml_tag}:{self.name}"



"""NetworkLink module."""

from typing import Iterable

from lxml import etree  # type: ignore

from pyLiveKML.objects.AbstractView import AbstractView
from pyLiveKML.objects.ExtendedData import ExtendedData
from pyLiveKML.objects.Feature import Feature
from pyLiveKML.objects.Link import Link
from pyLiveKML.objects.Object import _FieldDef, _ChildDef
from pyLiveKML.objects.Region import Region
from pyLiveKML.objects.StyleSelector import StyleSelector
from pyLiveKML.objects.TimePrimitive import TimePrimitive


class NetworkLink(Feature):
    """A KML 'NetworkLink' tag constructor.

    References a KML file or KMZ archive on a local or remote network. Use the `link`
    attribute to specify the location of the KML file. Within that `Link`, you can define
    the refresh options for updating the file, based on time and camera change.
    `NetworkLink`s can be used in combination with `Region`s to handle very large
    datasets efficiently.

    References
    ----------
    * https://developers.google.com/kml/documentation/kmlreference#networklink

    Parameters
    ----------
    name : str|None, default = None
        User-defined text displayed in the 3D viewer as the label for the object.
    visibility : bool | None, default = None
        Specifies whether the `Feature` is drawn in the 3D viewer when it is initially
        loaded. In order for a `Feature` to be visible, the `<visibility>` tag of all
        its ancestors must also be set `True`.
    is_open : bool | None, default = None
        Specifies whether a `Document` or `Folder` appears closed or open when first
        loaded into the "Places" panel. `False` or `None` is collapsed (the default),
        `True` is expanded. This element applies only to `Document`, `Folder`, and
        `NetworkLink`.
    author_name : str | None, default = None
        The name of the author of the `Feature`.
    author_link : str | None, default = None
        URL of the web page containing the KML file.
    address : str | None, default = None
        A string value representing an unstructured address written as a standard street,
        city, state address, and/or as a postal code.
    phone_number : str | None, default = None
        A string value representing a telephone number. This element is used by Google
        Maps Mobile only. The industry standard for Java-enabled cellular phones is
        RFC2806.
    snippet : str | None, default = None
        A short description of the `Feature`. In Google Earth, this description is
        displayed in the "Places" panel under the name of the `Feature`. If a `<Snippet>`
        is not supplied, the first two lines of the `<description>` are used. In Google
        Earth, if a `Placemark` contains both a `<description>` and a `<Snippet>`, the
        `<Snippet>` appears beneath the `Placemark` in the "Places" panel, and the
        `<description>` appears in the `Placemark`'s description balloon. This tag does
        not support HTML markup.
    snippet_max_lines : int | None, default = None
    description : str | None, default = None
        User-supplied content that appears in the description balloon. HTML *is*
        supported, but it is **highly** recommended to read the detailed documentation
        at
        https://developers.google.com/kml/documentation/kmlreference#elements-specific-to-feature
    abstract_view : AbstractView | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.AbstractView`, i.e. either a
        :class:`pyLiveKML.objects.Camera` or :class:`pyLiveKML.objects.LookAt`
    time_primitive : TimePrimitive | None, default = None
        Any concrete subclass of :class:`pyLiveKML.objects.TimePrimitive`, i.e. either a
        :class:`pyLiveKML.objects.TimeStamp` or :class:`pyLiveKML.objects.TimeSpan`
    style_url : str | None = None
        URL of a `<Style>` or `<StyleMap>` defined in a `<Document>`. If the style is in
        the same file, use a # reference. If the style is defined in an external file,
        use a full URL along with # referencing.
    styles : StyleSelector | Iterable[StyleSelector] | None, default = None
        One or more `Style`s and `StyleMap`s can be defined to customize the appearance
        of any element derived from `Feature` or of the `Geometry` in a `Placemark`. A
        style defined within a `Feature` is called an "inline style" and applies only to
        the `Feature` that contains it. A style defined as the child of a `<Document>` is
        called a "shared style." A shared style must have an id defined for it. This id
        is referenced by one or more `Features` within the `<Document>`. In cases where
        a style element is defined both in a shared style and in an inline style for a
        `Feature` — that is, a `Folder`, `GroundOverlay`, `NetworkLink`, `Placemark`, or
        `ScreenOverlay` — the value for the `Feature`'s inline style takes precedence over
        the value for the shared style.
    region : Region | None, default = None
        `Feature`s and `Geometry`'s associated with a `Region` are drawn only when the
        `Region` is active.
    extended_data : ExtendedData | None, default = None
        Allows you to add custom data to a KML file. This data can be:
        * Data that references an external XML schema.
        * Untyped data/value pairs.
        * Typed data.

        A given KML `Feature` can contain a combination of these types of custom data.
    fly_to_view: bool | None, default = None
        If `True`, causes Google Earth to fly to the view of the `LookAt` or `Camera` in
        the `NetworkLinkControl` (if it exists). If the `NetworkLinkControl` does not
        contain an `AbstractView` element, Google Earth flies to the `LookAt` or `Camera`
        element in the `Feature` child within the `<kml>` tag in the refreshed file. If the
        `<kml>` tag does not have a `LookAt` or `Camera` specified, or `fly_to_view` is
        `False`, the view is unchanged.
    refresh_visibility: bool | None, default = None
        If `False`, leaves the visibility of features within the control of the Google
        Earth user. Set to `True` to reset the visibility of features each time the
        `NetworkLink` is refreshed.
    link : Link | None, default = None
        Specifies the location of the KML file to be loaded.

    Attributes
    ----------
    Same as parameters.

    """

    _kml_tag = "NetworkLink"
    _kml_fields = Feature._kml_fields + (
        _FieldDef("fly_to_view", "flyToView"),
        _FieldDef("refresh_visibility", "refreshVisibility"),
    )
    _kml_children = Feature._kml_children + (_ChildDef("link"),)

    def __init__(
        self,
        # inherited from `Feature`
        name: str | None = None,
        visibility: bool | None = None,
        is_open: bool | None = None,
        author_name: str | None = None,
        author_link: str | None = None,
        address: str | None = None,
        phone_number: str | None = None,
        snippet: str | None = None,
        snippet_max_lines: int | None = None,
        description: str | None = None,
        abstract_view: AbstractView | None = None,
        time_primitive: TimePrimitive | None = None,
        style_url: str | None = None,
        styles: StyleSelector | Iterable[StyleSelector] | None = None,
        region: Region | None = None,
        extended_data: ExtendedData | None = None,
        # Specific to `NetworkLink`
        fly_to_view: bool | None = None,
        refresh_visibility: bool | None = None,
        link: Link | None = None,
    ):
        """NetworkLink instance constructor."""
        Feature.__init__(
            self,
            name=name,
            visibility=visibility,
            is_open=is_open,
            author_name=author_name,
            author_link=author_link,
            address=address,
            phone_number=phone_number,
            snippet=snippet,
            snippet_max_lines=snippet_max_lines,
            description=description,
            abstract_view=abstract_view,
            time_primitive=time_primitive,
            style_url=style_url,
            styles=styles,
            region=region,
            extended_data=extended_data,
        )
        self.fly_to_view = fly_to_view
        self.refresh_visibility = refresh_visibility
        self.link = link
